using System;
using System.Windows.Forms;
using System.Drawing;

namespace RFSExplorer {
	/// <summary>
	/// Summary description for WinFormUtils.
	/// </summary>
	public class WinFormUtils {
		/// <summary>
		/// Makes the text in the passed label fit it's width by trimming in the middle and replacing with ellipses.
		/// This should be called from the 'Paint' event of the label.
		/// </summary>
		/// <param name="lbl">Label with the text and the size.</param>
		/// <param name="graphics">System.Drawing.Graphics provided in the paint call (e.Graphics).</param>
		public void MakePathElliptical_OnPaint(Label lbl, Graphics graphics) {
			graphics.Clear(lbl.BackColor);
			RectangleF rect = new RectangleF(0, 0, lbl.Width, lbl.Height);
			Brush br = new SolidBrush(lbl.ForeColor);
			StringFormat sf = new StringFormat(StringFormatFlags.NoWrap);
			sf.Trimming = StringTrimming.EllipsisPath;
			graphics.DrawString(lbl.Text, lbl.Font, br, rect, sf);
		}
	}
}
