using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Diagnostics;
using System.Data;

using System.Text.RegularExpressions;
using System.Xml;
using System.Runtime.InteropServices;

namespace RFSExplorer {

	/// <summary>
	/// Summary description for frmMain.
	/// </summary>
	public class frmMain : System.Windows.Forms.Form {
		private Settings settings = null;
		private ColumnSorter listViewSorter;
		private MenuItem clonedViewMenu = null;

		#region ListView sort icon related code.
		[StructLayout(LayoutKind.Sequential)]
		private struct HDITEM {
			public Int32     mask;
			public Int32     cxy;
			[MarshalAs(UnmanagedType.LPTStr)]
			public String    pszText;
			public IntPtr	 hbm;
			public Int32     cchTextMax;
			public Int32     fmt;
			public Int32     lParam;
			public Int32     iImage;
			public Int32     iOrder;
		};

		[DllImport("user32")]
		static extern IntPtr SendMessage(IntPtr Handle, Int32 msg, IntPtr wParam, IntPtr lParam);

		[DllImport("user32", EntryPoint="SendMessage")]
		static extern IntPtr SendMessage2(IntPtr Handle, Int32 msg, IntPtr wParam, ref HDITEM lParam);

		#region const's
		const Int32 HDI_WIDTH			= 0x0001;
		const Int32 HDI_HEIGHT			= HDI_WIDTH;
		const Int32 HDI_TEXT			= 0x0002;
		const Int32 HDI_FORMAT			= 0x0004;
		const Int32 HDI_LPARAM			= 0x0008;
		const Int32 HDI_BITMAP			= 0x0010;
		const Int32 HDI_IMAGE			= 0x0020;
		const Int32 HDI_DI_SETITEM		= 0x0040;
		const Int32 HDI_ORDER			= 0x0080;
		const Int32 HDI_FILTER			= 0x0100;		// 0x0500

		const Int32 HDF_LEFT			= 0x0000;
		const Int32 HDF_RIGHT			= 0x0001;
		const Int32 HDF_CENTER			= 0x0002;
		const Int32 HDF_JUSTIFYMASK		= 0x0003;
		const Int32 HDF_RTLREADING		= 0x0004;
		const Int32 HDF_OWNERDRAW		= 0x8000;
		const Int32 HDF_STRING			= 0x4000;
		const Int32 HDF_BITMAP			= 0x2000;
		const Int32 HDF_BITMAP_ON_RIGHT = 0x1000;
		const Int32 HDF_IMAGE			= 0x0800;
		const Int32 HDF_SORTUP			= 0x0400;		// 0x0501
		const Int32 HDF_SORTDOWN		= 0x0200;		// 0x0501

		const Int32 LVM_FIRST           = 0x1000;		// List messages
		const Int32 LVM_GETHEADER		= LVM_FIRST + 31;

		const Int32 HDM_FIRST           = 0x1200;		// Header messages
		const Int32 HDM_SETIMAGELIST	= HDM_FIRST + 8;
		const Int32 HDM_GETIMAGELIST    = HDM_FIRST + 9;
		const Int32 HDM_GETITEM			= HDM_FIRST + 11;
		const Int32 HDM_SETITEM			= HDM_FIRST + 12;
		#endregion

		private void InitializeListViewIcons() {
			ImageList imgList = imgListView_Small;
			imgList.Images.Add(GetArrowBitmap(ArrowType.Ascending));		// Add ascending arrow
			imgList.Images.Add(GetArrowBitmap(ArrowType.Descending));		// Add descending arrow
		}

		private void ChangeHeaderIcon(int columnIndex, SortOrder sortOrder) {
			IntPtr hHeader = SendMessage(lvwMain.Handle, LVM_GETHEADER, IntPtr.Zero, IntPtr.Zero);
			HDITEM hd = new HDITEM();
			hd.mask = HDI_FORMAT | HDI_IMAGE;
			hd.fmt = HDF_LEFT | HDF_STRING | HDF_BITMAP_ON_RIGHT;

			if (sortOrder != SortOrder.None) {
				hd.fmt |= HDF_IMAGE;
				hd.iImage = (int)(sortOrder == SortOrder.Ascending ? ListViewImages.SortAscending : ListViewImages.SortDescending);
			}

			SendMessage2(hHeader, HDM_SETITEM, new IntPtr(columnIndex), ref hd);
		}

		enum ArrowType { Ascending, Descending }

		Bitmap GetArrowBitmap(ArrowType type) {
			Bitmap bmp = new Bitmap(24, 24);
			Graphics gfx = Graphics.FromImage(bmp);

			Pen lightPen = SystemPens.ControlLightLight;
			Pen shadowPen = SystemPens.ControlDark;

			gfx.FillRectangle(System.Drawing.Brushes.Magenta, 0, 0, 24, 24);

			if(type == ArrowType.Ascending) {
				gfx.DrawLine(lightPen, 8, 14, 15, 14);
				gfx.DrawLine(lightPen, 15, 14, 12, 7);
				gfx.DrawLine(shadowPen, 11, 7, 8, 14);
			}

			else if(type == ArrowType.Descending) {
				gfx.DrawLine(lightPen, 12, 14, 15, 7);
				gfx.DrawLine(shadowPen, 11, 14, 8, 7);
				gfx.DrawLine(shadowPen, 8, 7, 15, 7);
			}
			
			gfx.Dispose();

			return bmp;
		}

		private void SetHeaderImageList(ListView list, ImageList imgList) {
			IntPtr hHeader = SendMessage(list.Handle, LVM_GETHEADER, IntPtr.Zero, IntPtr.Zero);
			SendMessage(hHeader, HDM_SETIMAGELIST, IntPtr.Zero, imgList.Handle);
		}
		#endregion

		#region Controls
		private System.Windows.Forms.Button cmdRefreshTree;
		private System.Windows.Forms.MainMenu mnuMain;
		private System.Windows.Forms.MenuItem menuItem1;
		private System.Windows.Forms.MenuItem menuItem3;
		private System.Windows.Forms.TreeView tvwMain;
		private System.Windows.Forms.ListView lvwMain;
		private System.Windows.Forms.MenuItem menuItem5;
		private System.Windows.Forms.Button cmdTest;
		private System.Windows.Forms.MenuItem mnuTools_Settings;
		private System.Windows.Forms.ImageList imgTree;
		private System.Windows.Forms.ImageList imgListView_Small;
		private System.Windows.Forms.ImageList imgListView_Large;
		private System.Windows.Forms.MenuItem mnuView_Details;
		private System.Windows.Forms.MenuItem mnuView_LargeIcons;
		private System.Windows.Forms.MenuItem mnuView_SmallIcons;
		private System.Windows.Forms.MenuItem mnuFile_Exit;
		private System.Windows.Forms.ColumnHeader colName;
		private System.Windows.Forms.ColumnHeader colSize;
		private System.Windows.Forms.ColumnHeader colPermissions;
		private System.Windows.Forms.ColumnHeader colLastModified;
		private System.Windows.Forms.Panel pnlBottom;
		private System.Windows.Forms.StatusBar stbMain;
		private System.Windows.Forms.Panel pnlMiddle;
		private System.Windows.Forms.Panel pnlTop;
		private System.Windows.Forms.TextBox txtAddress;
		private System.Windows.Forms.Label lblAddress;
		private System.Windows.Forms.Button cmdGo;
		private System.Windows.Forms.StatusBarPanel stbpMain;
		private System.Windows.Forms.ContextMenu ctxmnuListView;
		private System.Windows.Forms.MenuItem mnuView;
		private System.Windows.Forms.StatusBarPanel stbpCount;
		private System.Windows.Forms.MenuItem mnuListView_Properties;
		private System.Windows.Forms.MenuItem mnuHelp_About;
		private System.Windows.Forms.Splitter spltMain;
		private System.Windows.Forms.MenuItem menuItem7;
		private System.ComponentModel.IContainer components;
		#endregion

		private enum TreeViewImages {
			Root,
			PartitionIcon,
			FolderClosed,
			FolderOpen
		}

		private enum ListViewImages {
			FolderClosed,
			UnknownDocument,
			SortAscending = 4,
			SortDescending = 5
		}

		public frmMain() {
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing ) {
			if( disposing ) {
				if (components != null) {
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent() {
			this.components = new System.ComponentModel.Container();
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(frmMain));
			this.cmdRefreshTree = new System.Windows.Forms.Button();
			this.mnuMain = new System.Windows.Forms.MainMenu();
			this.menuItem1 = new System.Windows.Forms.MenuItem();
			this.mnuFile_Exit = new System.Windows.Forms.MenuItem();
			this.mnuView = new System.Windows.Forms.MenuItem();
			this.mnuView_Details = new System.Windows.Forms.MenuItem();
			this.mnuView_LargeIcons = new System.Windows.Forms.MenuItem();
			this.mnuView_SmallIcons = new System.Windows.Forms.MenuItem();
			this.menuItem5 = new System.Windows.Forms.MenuItem();
			this.mnuTools_Settings = new System.Windows.Forms.MenuItem();
			this.menuItem3 = new System.Windows.Forms.MenuItem();
			this.mnuHelp_About = new System.Windows.Forms.MenuItem();
			this.pnlBottom = new System.Windows.Forms.Panel();
			this.cmdTest = new System.Windows.Forms.Button();
			this.pnlMiddle = new System.Windows.Forms.Panel();
			this.lvwMain = new System.Windows.Forms.ListView();
			this.colName = new System.Windows.Forms.ColumnHeader();
			this.colSize = new System.Windows.Forms.ColumnHeader();
			this.colPermissions = new System.Windows.Forms.ColumnHeader();
			this.colLastModified = new System.Windows.Forms.ColumnHeader();
			this.ctxmnuListView = new System.Windows.Forms.ContextMenu();
			this.menuItem7 = new System.Windows.Forms.MenuItem();
			this.mnuListView_Properties = new System.Windows.Forms.MenuItem();
			this.imgListView_Large = new System.Windows.Forms.ImageList(this.components);
			this.imgListView_Small = new System.Windows.Forms.ImageList(this.components);
			this.spltMain = new System.Windows.Forms.Splitter();
			this.tvwMain = new System.Windows.Forms.TreeView();
			this.imgTree = new System.Windows.Forms.ImageList(this.components);
			this.stbMain = new System.Windows.Forms.StatusBar();
			this.stbpMain = new System.Windows.Forms.StatusBarPanel();
			this.stbpCount = new System.Windows.Forms.StatusBarPanel();
			this.pnlTop = new System.Windows.Forms.Panel();
			this.cmdGo = new System.Windows.Forms.Button();
			this.lblAddress = new System.Windows.Forms.Label();
			this.txtAddress = new System.Windows.Forms.TextBox();
			this.pnlBottom.SuspendLayout();
			this.pnlMiddle.SuspendLayout();
			((System.ComponentModel.ISupportInitialize)(this.stbpMain)).BeginInit();
			((System.ComponentModel.ISupportInitialize)(this.stbpCount)).BeginInit();
			this.pnlTop.SuspendLayout();
			this.SuspendLayout();
			// 
			// cmdRefreshTree
			// 
			this.cmdRefreshTree.Location = new System.Drawing.Point(16, 8);
			this.cmdRefreshTree.Name = "cmdRefreshTree";
			this.cmdRefreshTree.Size = new System.Drawing.Size(184, 24);
			this.cmdRefreshTree.TabIndex = 1;
			this.cmdRefreshTree.Text = "&Refresh";
			this.cmdRefreshTree.Click += new System.EventHandler(this.cmdRefreshTree_Click);
			// 
			// mnuMain
			// 
			this.mnuMain.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					this.menuItem1,
																					this.mnuView,
																					this.menuItem5,
																					this.menuItem3});
			// 
			// menuItem1
			// 
			this.menuItem1.Index = 0;
			this.menuItem1.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					  this.mnuFile_Exit});
			this.menuItem1.Text = "&File";
			// 
			// mnuFile_Exit
			// 
			this.mnuFile_Exit.Index = 0;
			this.mnuFile_Exit.Text = "E&xit";
			this.mnuFile_Exit.Click += new System.EventHandler(this.mnuFile_Exit_Click);
			// 
			// mnuView
			// 
			this.mnuView.Index = 1;
			this.mnuView.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					this.mnuView_Details,
																					this.mnuView_LargeIcons,
																					this.mnuView_SmallIcons});
			this.mnuView.Text = "&View";
			// 
			// mnuView_Details
			// 
			this.mnuView_Details.Index = 0;
			this.mnuView_Details.Text = "&Details";
			this.mnuView_Details.Click += new System.EventHandler(this.mnuView_Details_Click);
			// 
			// mnuView_LargeIcons
			// 
			this.mnuView_LargeIcons.Checked = true;
			this.mnuView_LargeIcons.Index = 1;
			this.mnuView_LargeIcons.Text = "&Large Icons";
			this.mnuView_LargeIcons.Click += new System.EventHandler(this.mnuView_LargeIcons_Click);
			// 
			// mnuView_SmallIcons
			// 
			this.mnuView_SmallIcons.Index = 2;
			this.mnuView_SmallIcons.Text = "&Small Icons";
			this.mnuView_SmallIcons.Click += new System.EventHandler(this.mnuView_SmallIcons_Click);
			// 
			// menuItem5
			// 
			this.menuItem5.Index = 2;
			this.menuItem5.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					  this.mnuTools_Settings});
			this.menuItem5.Text = "&Tools";
			// 
			// mnuTools_Settings
			// 
			this.mnuTools_Settings.Index = 0;
			this.mnuTools_Settings.Text = "&Settings...";
			this.mnuTools_Settings.Click += new System.EventHandler(this.mnuTools_Settings_Click);
			// 
			// menuItem3
			// 
			this.menuItem3.Index = 3;
			this.menuItem3.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					  this.mnuHelp_About});
			this.menuItem3.Text = "&Help";
			// 
			// mnuHelp_About
			// 
			this.mnuHelp_About.Index = 0;
			this.mnuHelp_About.Text = "&About";
			this.mnuHelp_About.Click += new System.EventHandler(this.mnuHelp_About_Click);
			// 
			// pnlBottom
			// 
			this.pnlBottom.Controls.Add(this.cmdTest);
			this.pnlBottom.Controls.Add(this.cmdRefreshTree);
			this.pnlBottom.Dock = System.Windows.Forms.DockStyle.Bottom;
			this.pnlBottom.Location = new System.Drawing.Point(0, 368);
			this.pnlBottom.Name = "pnlBottom";
			this.pnlBottom.Size = new System.Drawing.Size(648, 48);
			this.pnlBottom.TabIndex = 6;
			this.pnlBottom.Visible = false;
			// 
			// cmdTest
			// 
			this.cmdTest.Location = new System.Drawing.Point(208, 8);
			this.cmdTest.Name = "cmdTest";
			this.cmdTest.Size = new System.Drawing.Size(120, 24);
			this.cmdTest.TabIndex = 2;
			this.cmdTest.Text = "T&est";
			// 
			// pnlMiddle
			// 
			this.pnlMiddle.Controls.Add(this.lvwMain);
			this.pnlMiddle.Controls.Add(this.spltMain);
			this.pnlMiddle.Controls.Add(this.tvwMain);
			this.pnlMiddle.Dock = System.Windows.Forms.DockStyle.Fill;
			this.pnlMiddle.DockPadding.All = 5;
			this.pnlMiddle.Location = new System.Drawing.Point(0, 32);
			this.pnlMiddle.Name = "pnlMiddle";
			this.pnlMiddle.Size = new System.Drawing.Size(648, 336);
			this.pnlMiddle.TabIndex = 7;
			// 
			// lvwMain
			// 
			this.lvwMain.Columns.AddRange(new System.Windows.Forms.ColumnHeader[] {
																					  this.colName,
																					  this.colSize,
																					  this.colPermissions,
																					  this.colLastModified});
			this.lvwMain.ContextMenu = this.ctxmnuListView;
			this.lvwMain.Dock = System.Windows.Forms.DockStyle.Fill;
			this.lvwMain.LargeImageList = this.imgListView_Large;
			this.lvwMain.Location = new System.Drawing.Point(237, 5);
			this.lvwMain.Name = "lvwMain";
			this.lvwMain.Size = new System.Drawing.Size(406, 326);
			this.lvwMain.SmallImageList = this.imgListView_Small;
			this.lvwMain.Sorting = System.Windows.Forms.SortOrder.Ascending;
			this.lvwMain.TabIndex = 4;
			this.lvwMain.DoubleClick += new System.EventHandler(this.lvwMain_DoubleClick);
			this.lvwMain.ColumnClick += new System.Windows.Forms.ColumnClickEventHandler(this.lvwMain_ColumnClick);
			this.lvwMain.ItemDrag += new System.Windows.Forms.ItemDragEventHandler(this.lvwMain_ItemDrag);
			this.lvwMain.SelectedIndexChanged += new System.EventHandler(this.lvwMain_SelectedIndexChanged);
			// 
			// colName
			// 
			this.colName.Text = "Name";
			// 
			// colSize
			// 
			this.colSize.Text = "Size";
			// 
			// colPermissions
			// 
			this.colPermissions.Text = "Permissions";
			// 
			// colLastModified
			// 
			this.colLastModified.Text = "Date Modified";
			// 
			// ctxmnuListView
			// 
			this.ctxmnuListView.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																						   this.menuItem7,
																						   this.mnuListView_Properties});
			this.ctxmnuListView.Popup += new System.EventHandler(this.ctxmnuListView_Popup);
			// 
			// menuItem7
			// 
			this.menuItem7.Index = 0;
			this.menuItem7.Text = "-";
			// 
			// mnuListView_Properties
			// 
			this.mnuListView_Properties.Index = 1;
			this.mnuListView_Properties.Text = "&Properties";
			this.mnuListView_Properties.Click += new System.EventHandler(this.mnuListView_Properties_Click);
			// 
			// imgListView_Large
			// 
			this.imgListView_Large.ColorDepth = System.Windows.Forms.ColorDepth.Depth24Bit;
			this.imgListView_Large.ImageSize = new System.Drawing.Size(48, 48);
			this.imgListView_Large.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("imgListView_Large.ImageStream")));
			this.imgListView_Large.TransparentColor = System.Drawing.Color.Black;
			// 
			// imgListView_Small
			// 
			this.imgListView_Small.ImageSize = new System.Drawing.Size(24, 24);
			this.imgListView_Small.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("imgListView_Small.ImageStream")));
			this.imgListView_Small.TransparentColor = System.Drawing.Color.Magenta;
			// 
			// spltMain
			// 
			this.spltMain.Location = new System.Drawing.Point(232, 5);
			this.spltMain.Name = "spltMain";
			this.spltMain.Size = new System.Drawing.Size(5, 326);
			this.spltMain.TabIndex = 5;
			this.spltMain.TabStop = false;
			// 
			// tvwMain
			// 
			this.tvwMain.Dock = System.Windows.Forms.DockStyle.Left;
			this.tvwMain.ImageList = this.imgTree;
			this.tvwMain.Location = new System.Drawing.Point(5, 5);
			this.tvwMain.Name = "tvwMain";
			this.tvwMain.Size = new System.Drawing.Size(227, 326);
			this.tvwMain.Sorted = true;
			this.tvwMain.TabIndex = 3;
			this.tvwMain.AfterExpand += new System.Windows.Forms.TreeViewEventHandler(this.tvwMain_AfterExpand);
			this.tvwMain.AfterSelect += new System.Windows.Forms.TreeViewEventHandler(this.tvwMain_AfterSelect);
			this.tvwMain.ItemDrag += new System.Windows.Forms.ItemDragEventHandler(this.tvwMain_ItemDrag);
			// 
			// imgTree
			// 
			this.imgTree.ImageSize = new System.Drawing.Size(16, 16);
			this.imgTree.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("imgTree.ImageStream")));
			this.imgTree.TransparentColor = System.Drawing.Color.Transparent;
			// 
			// stbMain
			// 
			this.stbMain.Location = new System.Drawing.Point(0, 416);
			this.stbMain.Name = "stbMain";
			this.stbMain.Panels.AddRange(new System.Windows.Forms.StatusBarPanel[] {
																					   this.stbpMain,
																					   this.stbpCount});
			this.stbMain.ShowPanels = true;
			this.stbMain.Size = new System.Drawing.Size(648, 22);
			this.stbMain.TabIndex = 8;
			// 
			// stbpMain
			// 
			this.stbpMain.AutoSize = System.Windows.Forms.StatusBarPanelAutoSize.Spring;
			this.stbpMain.Width = 532;
			// 
			// pnlTop
			// 
			this.pnlTop.Controls.Add(this.cmdGo);
			this.pnlTop.Controls.Add(this.lblAddress);
			this.pnlTop.Controls.Add(this.txtAddress);
			this.pnlTop.Dock = System.Windows.Forms.DockStyle.Top;
			this.pnlTop.Location = new System.Drawing.Point(0, 0);
			this.pnlTop.Name = "pnlTop";
			this.pnlTop.Size = new System.Drawing.Size(648, 32);
			this.pnlTop.TabIndex = 11;
			// 
			// cmdGo
			// 
			this.cmdGo.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Right)));
			this.cmdGo.Location = new System.Drawing.Point(568, 5);
			this.cmdGo.Name = "cmdGo";
			this.cmdGo.TabIndex = 2;
			this.cmdGo.Text = "&Go";
			this.cmdGo.Click += new System.EventHandler(this.cmdGo_Click);
			// 
			// lblAddress
			// 
			this.lblAddress.AutoSize = true;
			this.lblAddress.Location = new System.Drawing.Point(8, 8);
			this.lblAddress.Name = "lblAddress";
			this.lblAddress.Size = new System.Drawing.Size(49, 16);
			this.lblAddress.TabIndex = 0;
			this.lblAddress.Text = "A&ddress:";
			// 
			// txtAddress
			// 
			this.txtAddress.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.txtAddress.Location = new System.Drawing.Point(64, 6);
			this.txtAddress.Name = "txtAddress";
			this.txtAddress.Size = new System.Drawing.Size(488, 20);
			this.txtAddress.TabIndex = 1;
			this.txtAddress.Text = "";
			this.txtAddress.KeyPress += new System.Windows.Forms.KeyPressEventHandler(this.txtAddress_KeyPress);
			// 
			// frmMain
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(648, 438);
			this.Controls.Add(this.pnlMiddle);
			this.Controls.Add(this.pnlBottom);
			this.Controls.Add(this.stbMain);
			this.Controls.Add(this.pnlTop);
			this.Menu = this.mnuMain;
			this.Name = "frmMain";
			this.Text = "RFS Explorer";
			this.Closing += new System.ComponentModel.CancelEventHandler(this.frmMain_Closing);
			this.Load += new System.EventHandler(this.frmMain_Load);
			this.pnlBottom.ResumeLayout(false);
			this.pnlMiddle.ResumeLayout(false);
			((System.ComponentModel.ISupportInitialize)(this.stbpMain)).EndInit();
			((System.ComponentModel.ISupportInitialize)(this.stbpCount)).EndInit();
			this.pnlTop.ResumeLayout(false);
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main() {
			Application.Run(new frmMain());
		}

		/// <returns>An ArrayList of RFSObject's</returns>
		private ArrayList GetRFSPartitions() {
			Process p = new	Process();
			p.StartInfo = settings.GetRFSToolStartInfo();
			p.StartInfo.Arguments = " autodetect";
			p.Start();
			p.WaitForExit();
			string output = p.StandardOutput.ReadToEnd();

			Regex reg = new Regex(@"^Drive (?'Drive'\d+), Partition (?'Partition'\d+) is a ReiserFS\s+Partition Size: \d+ Bytes \((?'SizeInMB'[0-9.]+) MB\)", RegexOptions.Multiline);
			ArrayList rtn = new ArrayList();
			foreach (Match m in reg.Matches(output)) {
				int drive = int.Parse(m.Groups["Drive"].Value);
				int partition = int.Parse(m.Groups["Partition"].Value);
				double size = double.Parse(m.Groups["SizeInMB"].Value);
				string label = string.Format("{0}.{1} - {2} MB", drive, partition, size);
				RFSObject partInfo = new RFSObject(RFSObjectType.Directory, drive, partition, "/", label, "/");
				partInfo.Size = size;
				rtn.Add(partInfo);
			}

			return rtn;
		}


		private void frmMain_Load(object sender, System.EventArgs e) {
			InitializeListViewIcons();
			listViewSorter = new ColumnSorter(lvwMain);
			lvwMain.ListViewItemSorter = listViewSorter;
			lvwMain.Sorting = SortOrder.Ascending;

			settings =  Settings.GetInstance();
			string exePath = Application.ExecutablePath;
			string configFile = exePath.Substring(0, exePath.LastIndexOf("\\")) + "\\RFSExplorer.config";
			settings.SetConfigFile(configFile);

			if (!System.IO.File.Exists(configFile)) {
				ShowSettingsForm();
			}

			settings.LoadSettings(configFile);
			string error;
			if (!settings.AreRequiredSettingsValid(out error)) {
				MessageBox.Show("Invalid settings...\n\n\n" + error);
				ShowSettingsForm();
			}

			tvwMain.Width = settings.MainForm_TreeView_Width;
			colLastModified.Width = settings.MainForm_ListView_LastMod_Width;
			colPermissions.Width = settings.MainForm_ListView_Perm_Width;
			colName.Width = settings.MainForm_ListView_Name_Width;
			colSize.Width = settings.MainForm_ListView_Size_Width;
			lvwMain.Refresh();
			try {
				ChangeListView((View)Enum.Parse(typeof(View), settings.MainForm_ListView_View));
			} catch {} // Ignore any errors in conversion.
			
			clonedViewMenu = mnuView.CloneMenu();
			ctxmnuListView.MenuItems.Add(0, clonedViewMenu);

			LoadInitialTree();
		}


		private void LoadInitialTree() {
			string error;
			if (!settings.AreRequiredSettingsValid(out error)) {
				MessageBox.Show(error);
				return;
			}
			lvwMain.BeginUpdate();
			lvwMain.Items.Clear();
			lvwMain.EndUpdate();
			lvwMain.Sort();

			tvwMain.BeginUpdate();
			tvwMain.Nodes.Clear();
			TreeNode root = new TreeNode("My Computer");
			root.Tag = "/";
			root.SelectedImageIndex = root.ImageIndex = (int)TreeViewImages.Root;

			// Get a list of the local partitions.
			ArrayList RFSPartitions = GetRFSPartitions();
			TreeNode tn, dummyNode;
			foreach(RFSObject partInfo in RFSPartitions) {
				tn = root.Nodes.Add(partInfo.Label);
				tn.ImageIndex = tn.SelectedImageIndex = (int)TreeViewImages.PartitionIcon;
				dummyNode = tn.Nodes.Add("dummy");
				dummyNode.Tag = "dummy";

				tn.Tag = partInfo;
			}
			
			if (root.Nodes.Count == 0) {
				root.Nodes.Add("No RFS partitions were detected.");
			}

			tvwMain.Nodes.Add(root);
			tvwMain.EndUpdate();
		}

		private void cmdRefreshTree_Click(object sender, System.EventArgs e) {
			LoadInitialTree();
		}

		/// <summary>
		/// Loads the specified RFSObject. The specified object must have the absolute path, drive and partition numbers.
		/// </summary>
		/// <param name="path"></param>
		public void LoadDirectoryFromPath(RFSObject target) {
			if (target.ObjectType != RFSObjectType.Directory) { throw new Exception("The object must have a type of Directory."); }

			// First make sure that the tree is loaded.
			if (tvwMain.Nodes.Count == 0) { LoadInitialTree(); }

			// First find the correct root node.
			TreeNode curNode = null;
			tvwMain.Nodes[0].Expand();
			foreach(TreeNode childNode in tvwMain.Nodes[0].Nodes) {
				if (childNode.Tag == null) continue;
				RFSObject obj = (RFSObject) childNode.Tag;
				if (obj.DriveNumber == target.DriveNumber && obj.PartitionNumber == target.PartitionNumber) {
					curNode = childNode;
					break;
				}
			}
			if (curNode == null) { 
				stbpMain.Text = "Unable to find the root node.";
				return; 
			}
			curNode.Expand();

			// Now traverse down the directory path and make sure that the nodes are all loaded.
			string[] pieces = target.AbsolutePath.Split('/');
			foreach(string nextDir in pieces) {
				if (nextDir == "") continue;
				curNode.Expand();
				foreach(TreeNode childNode in curNode.Nodes) {
					if (childNode.Text == nextDir) {
						curNode = childNode;
						break;
					}
				}
				if (curNode.Text != nextDir) { // If the specified dir was not found.
					stbpMain.Text = string.Format("Unable to find \"{0}\" directory.", nextDir);
					return;
				}
			}
			curNode.Expand();
			tvwMain.SelectedNode = curNode;

			stbpMain.Text = "Finished loading " + ((RFSObject)curNode.Tag).AbsolutePath + ".";
		}

		/// <summary>
		/// Loads the tree view for the specified directory.
		/// </summary>
		private void LoadDirectory(TreeNode tn, bool loadListView) {
			if (Type.Equals(tn.Tag.GetType(), typeof(RFSObject))) {
				this.Cursor = Cursors.WaitCursor;
				RFSObject parentDir = (RFSObject) tn.Tag;
				bool isFirstDirectoryLoad = false;
				if (tn.Nodes.Count == 1 && tn.Nodes[0].Tag != null && tn.Nodes[0].Tag.ToString() == "dummy") {
					isFirstDirectoryLoad = true;
					tn.Nodes.Clear();
					tvwMain.BeginUpdate();
				}
				Process p = new Process();
				p.StartInfo = settings.GetRFSToolStartInfo();
				p.StartInfo.Arguments = "ls -p" + parentDir.Descriptor;
				p.Start();

				Regex reg = new Regex(@"^(?'Permissions'(?'SpecialBit'[cd-]{1})(?'AllPerms'[drwx-]{3})(?'GroupPerms'[drwx-]{3})(?'OwnerPerms'[drwx-]{3}))\s+(?'FileSizeOrDirInd'(<DIR>|\d+)) (?'ModDateTime'.{24}) (?'Name'.*?)\s*(\n|$)");
				if (loadListView) {
					txtAddress.Text = parentDir.Descriptor;
					lvwMain.BeginUpdate();
					lvwMain.Items.Clear();
				}
				string str;
				int objCount = 0;
				while ((str = p.StandardOutput.ReadLine()) != null) {
					foreach (Match m in reg.Matches(str)) {
						string name = m.Groups["Name"].Value.ToString();
						if (name == "." || name =="..") continue;
						objCount++;
						
						ListViewItem lvwItem = null;
						// If we are loading the list view then instantiate a new list view item.
						if (loadListView) {
							lvwItem = new ListViewItem();
							lvwItem.ImageIndex = (int)ListViewImages.UnknownDocument;
						}

						if (m.Groups["SpecialBit"].ToString() == "d") { // If this is a directory.
							TreeNode dir = null;
							if (isFirstDirectoryLoad) {
								dir = tn.Nodes.Add(name);
								TreeNode dummyNode = dir.Nodes.Add("dummy");
								dummyNode.Tag = "dummy";
							}

							RFSObject curDirObject = new RFSObject(RFSObjectType.Directory, parentDir, name, name, 
								m.Groups["ModDateTime"].Value.ToString(),
								m.Groups["Permissions"].Value.ToString());

							if (isFirstDirectoryLoad) {
								dir.Tag = curDirObject;
								dir.SelectedImageIndex = (int)TreeViewImages.FolderOpen;
								dir.ImageIndex = (int)TreeViewImages.FolderClosed;
							}
							if (loadListView) {
								lvwItem.Tag = curDirObject;
								lvwItem.ImageIndex = (int)ListViewImages.FolderClosed;
								lvwItem.SubItems.Add(curDirObject.GetHumanReadbleSize());
								lvwItem.SubItems.Add(curDirObject.Permissions);
								lvwItem.SubItems.Add(curDirObject.DateModified);
							}
						} else if (loadListView) { // If this is a file and we are refreshing the list view then add to the list view.
							RFSObject fileObject = new RFSObject(RFSObjectType.File, parentDir, name, name,
								m.Groups["ModDateTime"].Value.ToString(),
								m.Groups["Permissions"].Value.ToString());

							fileObject.Size = double.Parse(m.Groups["FileSizeOrDirInd"].Value);
							lvwItem.Tag = fileObject;
							lvwItem.SubItems.Add(fileObject.GetHumanReadbleSize());
							lvwItem.SubItems.Add(fileObject.Permissions);
							lvwItem.SubItems.Add(fileObject.DateModified);
						}

						if (loadListView) { 
							lvwItem.Text = name;
							lvwMain.Items.Add(lvwItem);
						}
					}
				}
				if (loadListView) {
					lvwMain.EndUpdate();
					stbpCount.Text = objCount + " objects";
				}
				//if (tn.Nodes.Count > 0 ) { tn.Expand(); }
				if (isFirstDirectoryLoad) { tvwMain.EndUpdate(); }
				this.Cursor = Cursors.Default;
			}
		}

		private void tvwMain_AfterSelect(object sender, System.Windows.Forms.TreeViewEventArgs e) {
			if (e.Node == null) return;
			LoadDirectory(e.Node, true);
		}

		private void tvwMain_AfterExpand(object sender, System.Windows.Forms.TreeViewEventArgs e) {
			if (e.Node == null) return;
			LoadDirectory(e.Node, false);
		}

		private void mnuTools_Settings_Click(object sender, System.EventArgs e) {
			ShowSettingsForm();
		}

		private void ShowSettingsForm() {
			frmSettings frm = new frmSettings();
			frm.ShowDialog();
			LoadInitialTree();
		}

		private void mnuView_Details_Click(object sender, System.EventArgs e) {
			ChangeListView(View.Details);
		}

		private void mnuView_LargeIcons_Click(object sender, System.EventArgs e) {
			ChangeListView(View.LargeIcon);
		}

		private void mnuView_SmallIcons_Click(object sender, System.EventArgs e) {
			ChangeListView(View.SmallIcon);
		}

		private void ChangeListView(View newView) {
			mnuView_Details.Checked = newView == View.Details;
			mnuView_SmallIcons.Checked = newView == View.SmallIcon;
			mnuView_LargeIcons.Checked = newView == View.LargeIcon;

			if (clonedViewMenu != null) {
				clonedViewMenu.MenuItems[0].Checked = newView == View.Details;
				clonedViewMenu.MenuItems[1].Checked = newView == View.LargeIcon;
				clonedViewMenu.MenuItems[2].Checked = newView == View.SmallIcon;
			}

			lvwMain.View = newView;
		}

		private void mnuFile_Exit_Click(object sender, System.EventArgs e) {
			this.Close();
		}

		private void lvwMain_ColumnClick(object sender, System.Windows.Forms.ColumnClickEventArgs e) {
			ChangeHeaderIcon(listViewSorter.CurrentColumn, SortOrder.None); // Clear the last icon.

			listViewSorter.CurrentColumn = e.Column;
			lvwMain.Sorting = (lvwMain.Sorting == SortOrder.Descending ? SortOrder.Ascending : SortOrder.Descending);
			lvwMain.Sort();

			ChangeHeaderIcon(e.Column, lvwMain.Sorting);
		}

		private void lvwMain_ItemDrag(object sender, System.Windows.Forms.ItemDragEventArgs e) {
			ArrayList objectsToCopy = new ArrayList();
			foreach (ListViewItem lvwItem in lvwMain.SelectedItems) {
				if (lvwItem.Tag != null) { objectsToCopy.Add(lvwItem.Tag); }
			}

			frmCopy frm = new frmCopy(objectsToCopy);
			frm.ShowDialog();
		}

		private void tvwMain_ItemDrag(object sender, System.Windows.Forms.ItemDragEventArgs e) {
			TreeNode tn = (TreeNode) e.Item;
			if (tn.Tag == null || !tn.Tag.GetType().Equals(typeof(RFSObject))) { return; }

			if (((RFSObject)tn.Tag).AbsolutePath == "/") { MessageBox.Show("Copying of root is not allowed. Please copy individual objects."); return; }

			ArrayList objectsToCopy = new ArrayList();
			objectsToCopy.Add(((TreeNode)e.Item).Tag);

			frmCopy frm = new frmCopy(objectsToCopy);
			frm.ShowDialog();
		}

		private void ctxmnuListView_Popup(object sender, System.EventArgs e) {
			if (lvwMain.SelectedItems.Count == 0) {
				mnuListView_Properties.Enabled = false;
			} else {
				mnuListView_Properties.Enabled = true;
			}
		}

		private void mnuHelp_About_Click(object sender, System.EventArgs e) {
			frmAbout frm = new frmAbout();
			frm.ShowDialog();
		}

		private void frmMain_Closing(object sender, System.ComponentModel.CancelEventArgs e) {
			settings.MainForm_TreeView_Width = tvwMain.Width;
			settings.MainForm_ListView_LastMod_Width = colLastModified.Width;
			settings.MainForm_ListView_Perm_Width = colPermissions.Width;
			settings.MainForm_ListView_Size_Width = colSize.Width;
			settings.MainForm_ListView_Name_Width = colName.Width;
			settings.MainForm_ListView_View = lvwMain.View.ToString();
			settings.WriteSettings();
		}

		private void mnuListView_Properties_Click(object sender, System.EventArgs e) {
			frmProperties frm = new frmProperties((RFSObject) lvwMain.SelectedItems[0].Tag);
			frm.ShowDialog();
		}

		private void lvwMain_DoubleClick(object sender, System.EventArgs e) {
			if (lvwMain.SelectedItems.Count == 1 && lvwMain.SelectedItems[0].Tag.GetType().Equals(typeof(RFSObject))) {
				RFSObject obj = (RFSObject)lvwMain.SelectedItems[0].Tag;
				if (obj.ObjectType != RFSObjectType.Directory) { return; }
				LoadDirectoryFromPath(obj);
			}
		}

		private void lvwMain_SelectedIndexChanged(object sender, System.EventArgs e) {
		
		}

		private void cmdGo_Click(object sender, System.EventArgs e) {
			Regex regex = new Regex(@"^(?'DriveNumber'\d+)\.(?'PartitionNumber'\d+)\s+\""?(?'Path'(\w|[/])+)\""?$");
			if (!regex.IsMatch(txtAddress.Text)) {
				MessageBox.Show("The path must be in the format: driveNumber.partitionNumber path.\n\nFor example:\n\t0.16 /etc/X11","Invalid Format");
				return;
			}

			Match m = regex.Match(txtAddress.Text);
			if (m.Groups.Count == 0) { 
				MessageBox.Show("Sorry, unable to find any regex matches. Please inform Farhan of what you entered.", 
					"Unexpected Error", MessageBoxButtons.OK, MessageBoxIcon.Error); 
				return;
			}

			RFSObject obj = new RFSObject(RFSObjectType.Directory, 
				int.Parse(m.Groups["DriveNumber"].Value),
				int.Parse(m.Groups["PartitionNumber"].Value),
				m.Groups["Path"].Value);
			LoadDirectoryFromPath(obj);
		}

		private void txtAddress_KeyPress(object sender, System.Windows.Forms.KeyPressEventArgs e) {
			if(e.KeyChar == (char)13) { e.Handled = true; cmdGo_Click(null, null); }
		}
	}

	public  class  ColumnSorter : IComparer {
		public int CurrentColumn = 0;
		public ListView ListViewControl = null;
		/// <returns>
		/// Less than zero		x is less than y.
		/// Zero				x equals y.
		/// Greater than zero	x is greater than y.
		/// </returns>
		public int Compare(object x, object y) {
			if (ListViewControl.Sorting == SortOrder.None) { return 0; }

			int result = -100;
			ListViewItem rowA = (ListViewItem)x;
			ListViewItem rowB = (ListViewItem)y;

			RFSObject objA = (RFSObject)rowA.Tag;
			RFSObject objB = (RFSObject)rowB.Tag;

			if (CurrentColumn == 0) {
				result = objA.CompareTo(objB);
			} else if (CurrentColumn == 1) {
				result = objA.Size.CompareTo(objB.Size);
			} else if (CurrentColumn == 2) {
				result = objA.Permissions.CompareTo(objB.Permissions);
			} else if (CurrentColumn == 3) {
				result = objA.DateModified.CompareTo(objB.DateModified);
			}

			if (result == -100) {
				throw new ArgumentException("This shouldn't happen!");
			} else {
				return (ListViewControl.Sorting == SortOrder.Descending ? -1 : 1) * result;
			}
		}

		public ColumnSorter(ListView lvw) { ListViewControl = lvw; }
	}
}
